"""
Unit tests for pharokka overall
Usage: pytest .

"""

# import
import os
import shutil
# import functions
import subprocess
import sys
import unittest
from pathlib import Path
from unittest.mock import patch

import pytest
from loguru import logger

from bin.util import remove_directory

# import functions


# test data
test_data = Path("tests/test_data")
functions_data = Path(f"{test_data}/functions_files")
database_dir = Path(f"{test_data}/database")
overall_data = Path(f"{test_data}/overall")
meta_data = Path(f"{overall_data}/Meta_example")
rv_data = Path(f"{overall_data}/rv_example")
bug_data = Path(f"{overall_data}/bug_examples")
standard_data = Path(f"{overall_data}/Standard_examples")
standard_data_output = Path(f"{standard_data}/SAOMS1_Output")
stop_recoding_data = Path(f"{overall_data}/stop_recoding")
custom_db = Path(f"{test_data}/custom_db/microvirus.h3m")
custom_data = Path(f"{overall_data}/custom_examples")
tmrna_data = Path(f"{overall_data}/tmRNA_example")
AMR_data = Path(f"{overall_data}/AMR_example")
CRISPR_data = Path(f"{overall_data}/CRISPR_example")
VFDB_data = Path(f"{overall_data}/VFDB_example")
genbank_data = Path(f"{overall_data}/genbank_examples")
logger.add(lambda _: sys.exit(1), level="ERROR")
threads = 8


def remove_directory(dir_path):
    if os.path.exists(dir_path):
        shutil.rmtree(dir_path)


@pytest.fixture(scope="session")
def tmp_dir(tmpdir_factory):
    return tmpdir_factory.mktemp("tmp")


temp_dir = Path(f"{test_data}/fake_out")


def exec_command(cmnd, stdout=subprocess.PIPE, stderr=subprocess.PIPE):
    """executes shell command and returns stdout if completes exit code 0
    Parameters
    ----------
    cmnd : str
      shell command to be executed
    stdout, stderr : streams
      Default value (PIPE) intercepts process output, setting to None
      blocks this."""

    proc = subprocess.Popen(cmnd, shell=True, stdout=stdout, stderr=stderr)
    out, err = proc.communicate()
    if proc.returncode != 0:
        raise RuntimeError(f"FAILED: {cmnd}\n{err}")
    return out.decode("utf8") if out is not None else None


def test_download(tmp_dir):
    """test pharokka download"""
    cmd = f"install_databases.py -o {database_dir}"
    exec_command(cmd)


def test_overall(tmp_dir):
    """test pharokka overall"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f"
    exec_command(cmd)

def test_overall_reverse_mmseqs_sensitivity(tmp_dir):
    """test pharokka reverse_mmseqs - use with low sensitivity or else takes forever"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    sensitivity = "0.5"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --reverse_mmseqs2 -g pyrodigal-rv --sensitivity {sensitivity}"
    exec_command(cmd)

def test_overall_trna_anticodon(tmp_dir):
    """test for #405 trna anticodon bug"""
    input_fasta: Path = f"{bug_data}/AJ251789_trna_anticodon.fa"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --fast"
    exec_command(cmd)

def test_overall_pyhmmer_alphabet(tmp_dir):
    """test for #331 #357 pyhmmer alphabet must be explictly specifyied as amino acid"""
    input_fasta: Path = f"{bug_data}/pyhmmer_alphabet_issue357.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --fast"
    exec_command(cmd)

def test_terminase(tmp_dir):
    """test pharokka vfdb issue #410 fix with extra [] in vfdb headers"""
    input_fasta: Path = f"{bug_data}/vfdb_issue_410.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --fast"
    exec_command(cmd)

def test_overall_mash_distance(tmp_dir):
    """test pharokka overall with stricter mash distance"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --mash_distance 0.05"
    exec_command(cmd)


def test_overall_crispr(tmp_dir):
    """test pharokka overall crispr"""
    input_fasta: Path = f"{CRISPR_data}/Biggiephage_A_fullcontig_CasΦ1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f"
    exec_command(cmd)


def test_overall_crispr_minced_args(tmp_dir):
    """test pharokka crispr with minced args"""
    input_fasta: Path = f"{CRISPR_data}/Biggiephage_A_fullcontig_CasΦ1.fasta"
    cmd = f'pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -g prodigal --minced_args "minNR 2 -minRL 21" '
    exec_command(cmd)


def test_overall_vfdb(tmp_dir):
    """test pharokka overall on a phage with vfdb hits. Also include --skip_extra_annotations"""
    input_fasta: Path = f"{VFDB_data}/NC_004617.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --skip_extra_annotations"
    exec_command(cmd)


def test_overall_amr(tmp_dir):
    """test pharokka overall amr also includes '--skip_mash"""
    input_fasta: Path = f"{AMR_data}/NC_007458.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --skip_mash"
    exec_command(cmd)


def test_overall_tmrna(tmp_dir):
    """test pharokka overall tmrna"""
    input_fasta: Path = f"{tmrna_data}/NC_051700.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f"
    exec_command(cmd)


def test_overall_numeric_header_prodigal(tmp_dir):
    """#317 test pharokka prodigal with  numeric header #334"""
    input_fasta: Path = f"{standard_data}/SAOMS1_numeric_header.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -g prodigal --fast -m"
    exec_command(cmd)


def test_overall_numeric_header_phanotate(tmp_dir):
    """#317 test pharokka phanotate  with numeric header #334"""
    input_fasta: Path = f"{standard_data}/SAOMS1_numeric_header.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --fast "
    exec_command(cmd)


def test_overall_numeric_header_prodigal_gv(tmp_dir):
    """#317 test pharokka prodigal-gv with numeric header #334"""
    input_fasta: Path = f"{standard_data}/SAOMS1_numeric_header.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -m --fast"
    exec_command(cmd)


def test_meta(tmp_dir):
    """test pharokka meta"""
    input_fasta: Path = f"{meta_data}/combined_meta.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -m"
    exec_command(cmd)


def test_meta_unicycler_header_prodigal(tmp_dir):
    """#317 test pharokka overall with unicycler header - with all trna, tmras crispr etc"""
    input_fasta: Path = f"{meta_data}/combined_meta_unicycler_headers.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -g prodigal-gv -m"
    exec_command(cmd)

def test_pyrodigal_gv(tmp_dir):
    """test pharokka with prodigal-rv"""
    input_fasta: Path = f"{meta_data}/combined_meta.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -m -g prodigal-gv"
    exec_command(cmd)


def test_prodigal_rv(tmp_dir):
    """test pharokka with prodigal-rv"""
    input_fasta: Path = f"{rv_data}/Tymoviridae.fna"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -g pyrodigal-rv"
    exec_command(cmd)


def test_meta_dnaapler_all_bug(tmp_dir):
    """test pharokka meta dnaapler bug and split"""
    input_fasta: Path = f"{meta_data}/combined_meta.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -m -s --dnaapler --meta_hmm"
    exec_command(cmd)


def test_overall_locus(tmp_dir):
    """test pharokka overall locus tag prefix"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -l SAOMS1 -p SAOMS1"
    exec_command(cmd)


def test_custom(tmp_dir):
    """test pharokka overall with custom db"""
    input_fasta: Path = f"{custom_data}/MH649026.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} --custom_hmm {custom_db} -o {tmp_dir} -t {threads} -f "
    exec_command(cmd)


def test_custom_meta(tmp_dir):
    """test pharokka overall with custom db"""
    input_fasta: Path = f"{custom_data}/hundred_microviruses.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} --custom_hmm {custom_db} -o {tmp_dir} -t {threads} -f -m"
    exec_command(cmd)


def test_overall_prodigal(tmp_dir):
    """test pharokka overall prodigal"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -g prodigal"
    exec_command(cmd)


def test_overall_stop_recode(tmp_dir):
    """test pharokka overall recoded"""
    input_fasta: Path = f"{stop_recoding_data}/table_4/SRR1747055_scaffold_7.fa"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -g prodigal -c 4"
    exec_command(cmd)


def test_overall_dnaapler(tmp_dir):
    """test pharokka overall dnaapler"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --dnaapler"
    exec_command(cmd)


def test_overall_fast(tmp_dir):
    """test pharokka overall fast"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --fast"
    exec_command(cmd)


def test_overall_mmseqs(tmp_dir):
    """test pharokka overall mmseqs2_only"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --mmseqs2_only"
    exec_command(cmd)


def test_meta_no_cds_contig(tmp_dir):
    """test pharokka meta with a contig with empty contigs"""
    input_fasta: Path = f"{meta_data}/fake_meta.fa"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -m"
    exec_command(cmd)


######
# pharokka CI was timing out (>6 hours)
# These are covered by other rules anyway
# so just run as is

# def test_meta_hmm(tmp_dir):
#     """test pharokka meta hmm"""
#     input_fasta: Path = f"{meta_data}/fake_meta.fa"
#     cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -m --meta_hmm"
#     exec_command(cmd)


# def test_meta_split(tmp_dir):
#     """test pharokka meta split"""
#     input_fasta: Path = f"{meta_data}/fake_meta.fa"
#     cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f -m -s"
#     exec_command(cmd)


def test_terminase(tmp_dir):
    """test pharokka terminase"""
    input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
    cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {tmp_dir} -t {threads} -f --terminase --terminase_start 340 --terminase_strand neg"
    exec_command(cmd)


def test_overall_genbank(tmp_dir):
    """test pharokka overall with genbank input"""
    input_gbk: Path = f"{genbank_data}/SAOMS1.gbk"
    cmd = f"pharokka.py -i {input_gbk} -d {database_dir} -o {tmp_dir} -t {threads} -f --genbank"
    exec_command(cmd)


def test_overall_genbank_meta(tmp_dir):
    """test pharokka overall meta with genbank input"""
    input_gbk: Path = f"{genbank_data}/hundred_microviruses.gbk"
    cmd = f"pharokka.py -i {input_gbk} -d {database_dir} -o {tmp_dir} -t {threads} -f --genbank -m --meta_hmm"
    exec_command(cmd)


class testFails(unittest.TestCase):
    """Tests for fails"""

    def test_dupe_header(self):
        """tests that pharokka exits if a duplicate header is passed"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/dupe_header.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t 1 -f -m"
            exec_command(cmd)

    def test_overall_hash_header(self):
        """test pharokka overall with # in header - should error out"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1_hash_header.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t {threads} -f"
            exec_command(cmd)

    def test_meta_with_single_contig(self):
        """tests that pharokka exits if single contig is passed to meta"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t 1 -f -m"
            exec_command(cmd)

    def test_meta_terminas(self):
        """tests that pharokka exits if multiple contigs passed to meta"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{meta_data}/fake_meta.fa"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t 1 -f -m --terminase"
            exec_command(cmd)

    def test_bad_terminase(self):
        """tests that pharokka exits if only terminase specified"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t 1 -f --terminase"
            exec_command(cmd)

    def test_bad_terminase_start(self):
        """tests that pharokka exits if bad terminase start specified"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t 1 -f --terminase --terminase_start sf"
            exec_command(cmd)

    def test_bad_terminase_strand(self):
        """tests that pharokka exits if bad strand"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t 1 -f --terminase --terminase_start 34 --terminase_strand posit"
            exec_command(cmd)

    def test_bad_threads(self):
        """tests that pharokka exits if bad threads"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = (
                f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t sf -f"
            )
            exec_command(cmd)

    def test_bad_fast_mmseqs2_only(self):
        """tests that pharokka exits if both fast and mmseqs2_only"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t {threads} --fast --mmseqs2_only -f"
            exec_command(cmd)

    def test_bad_fast_meta_gmm(self):
        """tests that pharokka exits if both fast and meta_hmm"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t {threads} --fast --meta_hmm -m -f"
            exec_command(cmd)

    def test_bad_fast_meta_gmm(self):
        """tests that pharokka exits if both mmseqs2_only and meta_hmm"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t {threads} --mmseqs2_only --meta_hmm -m -f"
            exec_command(cmd)

    def test_bad_gene_pred(self):
        """tests that pharokka exits if bad gene predictior"""
        with self.assertRaises(RuntimeError):
            input_fasta: Path = f"{standard_data}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} -o {temp_dir} -t 1 -g proddigal -f"
            exec_command(cmd)

    def test_bad_genbank(self):
        """test pharokka overall with genbank input but no --genbank"""
        with self.assertRaises(RuntimeError):
            input_gbk: Path = f"{standard_data_output}/SAOMS1.fasta"
            cmd = f"pharokka.py -i {input_gbk} -d {database_dir} -o {temp_dir} -t {threads} -f "
            exec_command(cmd)

    def test_custom_bad(self):
        """test pharokka overall with bad custom db"""
        with self.assertRaises(RuntimeError):
            custom_db: Path = f"{standard_data_output}/SAOMS1.fasta"
            input_fasta: Path = f"{custom_data}/MH649026.fasta"
            cmd = f"pharokka.py -i {input_fasta} -d {database_dir} --custom_hmm {custom_db} -o {temp_dir} -t {threads} -f "
            exec_command(cmd)


remove_directory(temp_dir)
remove_directory(f"{database_dir}")



import argparse
import os
import shutil
import subprocess as sp
from argparse import RawTextHelpFormatter

import pyrodigal
import pyrodigal_gv
import pyrodigal_rv
from Bio import SeqIO
from loguru import logger
from util import get_version


def get_input():
    parser = argparse.ArgumentParser(
        description="pharokka: fast phage annotation program",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "-i", "--infile", action="store", help="Input genome file in fasta format."
    )
    parser.add_argument(
        "-o",
        "--outdir",
        action="store",
        help="Directory to write the output to.",
        default=os.path.join(os.getcwd(), "output/"),
    )
    parser.add_argument(
        "-d",
        "--database",
        action="store",
        help="Database directory. If the databases have been installed in the default directory, this is not required. Otherwise specify the path.",
        default="Default",
    )
    parser.add_argument(
        "-t",
        "--threads",
        help="Number of threads. Defaults to 1.",
        action="store",
        default=str(1),
    )
    parser.add_argument(
        "-f", "--force", help="Overwrites the output directory.", action="store_true"
    )
    parser.add_argument(
        "-p",
        "--prefix",
        action="store",
        help="Prefix for output files. This is not required.",
        default="Default",
    )
    parser.add_argument(
        "-l",
        "--locustag",
        action="store",
        help="User specified locus tag for the gff/gbk files. This is not required. A random locus tag will be generated instead.",
        default="Default",
    )
    parser.add_argument(
        "-g",
        "--gene_predictor",
        action="store",
        help='User specified gene predictor. Use "-g phanotate" or "-g prodigal" or "-g prodigal-gv" or "-g pyrodigal-rv" or "-g genbank". \nDefaults to phanotate usually and prodigal-gv in meta mode.',
        default="default",
    )
    parser.add_argument(
        "-m",
        "--meta",
        help="meta mode for metavirome input samples",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--split",
        help="split mode for metavirome samples. -m must also be specified. \nWill output separate split FASTA, gff and genbank files for each input contig.",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--coding_table",
        help="translation table for prodigal. Defaults to 11.",
        action="store",
        default="11",
    )
    parser.add_argument(
        "-e",
        "--evalue",
        help="E-value threshold for MMseqs2 database PHROGs, VFDB and CARD and PyHMMER PHROGs database search. Defaults to 1E-05.",
        action="store",
        default="1E-05",
    )
    parser.add_argument(
        "--fast",
        "--hmm_only",
        help="Runs PyHMMER (HMMs) with PHROGs only, not MMseqs2 with PHROGs, CARD or VFDB. \nDesigned for phage isolates, will not likely be faster for large metagenomes.",
        action="store_true",
    )
    parser.add_argument(
        "--mmseqs2_only",
        help="Runs MMseqs2 with PHROGs, CARD and VFDB only (same as Pharokka v1.3.2 and prior). Default in meta mode.",
        action="store_true",
    )
    parser.add_argument(
        "--meta_hmm",
        help="Overrides --mmseqs2_only in meta mode. Will run both MMseqs2 and PyHMMER.",
        action="store_true",
    )
    parser.add_argument(
        "--dnaapler",
        help="Runs dnaapler to automatically re-orient all contigs to begin with terminase large subunit if found. \nRecommended over using '--terminase'.",
        action="store_true",
    )
    parser.add_argument(
        "--custom_hmm",
        help="Run pharokka with a custom HMM profile database suffixed .h3m. \nPlease use create this with the create_custom_hmm.py script.",
        action="store",
        default="",
    )
    parser.add_argument(
        "--genbank",
        help="Flag denoting that -i/--input is a genbank file instead of the usual FASTA file. \n The CDS calls in this file will be preserved and re-annotated.",
        action="store_true",
    )
    parser.add_argument(
        "--terminase",
        help="Runs terminase large subunit re-orientation mode. \nSingle genome input only and requires --terminase_strand and --terminase_start to be specified.",
        action="store_true",
    )
    parser.add_argument(
        "--terminase_strand",
        help='Strand of terminase large subunit. Must be "pos" or "neg".',
        action="store",
        default="nothing",
    )
    parser.add_argument(
        "--terminase_start",
        help="Start coordinate of the terminase large subunit.",
        action="store",
        default="nothing",
    )
    parser.add_argument(
        "--skip_extra_annotations",
        help="Skips tRNAscan-SE 2, MinCED and Aragorn.",
        action="store_true",
    ),
    parser.add_argument(
        "--skip_mash",
        help="Skips running mash to find the closest match for each contig in INPHARED.",
        action="store_true",
    )
    parser.add_argument(
        "--minced_args",
        help='extra commands to pass to MINced (please omit the leading hyphen for the first argument). You will need to use quotation marks e.g. --minced_args "minNR 2 -minRL 21"',
        default="",
        type=str,
    )
    parser.add_argument(
        "--mash_distance",
        help="mash distance for the search against INPHARED. Defaults to 0.2.",
        default=0.2,
        type=float,
    )
    parser.add_argument(
        "--trna_scan_model",
        help="tRNAscan-SE model",
        choices=["general", "bacterial"],
        default="general",
        type=str,
    ),
    parser.add_argument(
        "--keep_raw_prodigal",
        help="Keeps raw prodigal header information.",
        action="store_true",
    )
    parser.add_argument(
        "--reverse_mmseqs2",
        help="MMseqs2 database as target not query.",
        action="store_true",
    )
    parser.add_argument(
        "--sensitivity",
        help="MMseqs2 sensitivity.",
        default=8.5,
        type=float,
    )
    parser.add_argument(
        "-V",
        "--version",
        help="Print pharokka Version",
        action="version",
        version=get_version(),
    )
    parser.add_argument(
        "--citation", help="Print pharokka Citation", action="store_true"
    )
    args = parser.parse_args()

    return args


def instantiate_dirs(output_dir, meta, force):
    # remove outdir on force
    if force == True:
        if os.path.isdir(output_dir) == True:
            logger.info(
                f"Removing output directory {output_dir} as -f or --force was specified."
            )
            shutil.rmtree(output_dir)

        elif os.path.isfile(output_dir) == True:
            os.remove(output_dir)
        else:
            logger.info(
                f"--force was specified even though the output directory {output_dir} does not already exist. Continuing."
            )
    else:
        if os.path.isdir(output_dir) == True or os.path.isfile(output_dir) == True:
            logger.error(
                f"The output directory {output_dir} already exists and force was not specified. Please specify -f or --force to overwrite it."
            )

    # instantiate outdir
    if os.path.isdir(output_dir) == False:
        os.mkdir(output_dir)
    mmseqs_dir = os.path.join(output_dir, "mmseqs/")
    if os.path.isdir(mmseqs_dir) == False:
        os.mkdir(mmseqs_dir)
    vfdb_dir = os.path.join(output_dir, "VFDB/")
    if os.path.isdir(vfdb_dir) == False:
        os.mkdir(vfdb_dir)
    CARD_dir = os.path.join(output_dir, "CARD/")
    if os.path.isdir(CARD_dir) == False:
        os.mkdir(CARD_dir)

    # tmp dir for  meta mode trnascan and phanotate
    input_tmp_dir = os.path.join(output_dir, "input_split_tmp/")
    if meta == True:
        if os.path.isdir(input_tmp_dir) == False:
            os.mkdir(input_tmp_dir)

    return output_dir


def validate_fasta(filename):
    if os.path.isfile(filename) == False:  # if file doesnt exist
        logger.error(
            f"Error: Input file {filename} does not exist. Please check your input."
        )
    else:
        with open(filename, "r") as handle:
            fasta = SeqIO.parse(handle, "fasta")
            logger.info(f"Checking input {filename}.")
            if any(fasta):
                logger.info(f"Input {filename} is in FASTA format.")
            else:
                logger.error("Error: Input file is not in the FASTA format.")

    # check for duplicate headers
    logger.info(f"Checking input {filename} for duplicate FASTA headers.")
    check_duplicate_headers(filename)
    logger.info(f"All headers in {filename} are unique.")


def check_duplicate_headers(fasta_file):
    """
    checks if there are duplicated in the FASTA header
    in response to Tina's issue
    https://github.com/gbouras13/pharokka/issues/293
    """
    header_set = set()

    # Iterate through the FASTA file and check for duplicate headers
    for record in SeqIO.parse(fasta_file, "fasta"):
        header = record.description
        if header in header_set:
            logger.error(
                f"Duplicate contig header found: {header}"
            )  # errors if duplicate header found
        if "#" in header:
            logger.error(
                f"# character found in contig: {header} - please remove all '#' from your contig headers"
            )  # errors if duplicate header found
        else:
            header_set.add(header)
    # if it finished it will be fine


def validate_gene_predictor(gene_predictor, genbank_flag):
    if gene_predictor == "phanotate":
        logger.info("Phanotate will be used for gene prediction.")
    elif gene_predictor == "prodigal":
        logger.info(
            "Prodigal implemented with pyrodigal will be used for gene prediction."
        )
    elif gene_predictor == "prodigal-gv":
        logger.info(
            "Prodigal-gv implemented with pyrodigal-gv will be used for gene prediction."
        )
    elif gene_predictor == "pyrodigal-rv":
        logger.info(
            "pyrodigal-rv will be used for gene prediction."
        )
    elif gene_predictor == "genbank":
        if genbank_flag is False:
            logger.error(
                "Error: you have specified -g genbank without --genbank. Please just use --genbank."
            )
    else:
        logger.error(
            "Error: gene predictor was incorrectly specified. Please use 'phanotate', 'prodigal', 'prodigal-gv' or 'pyrodigal-rv'."
        )


def validate_meta(filepath_in, meta, split, genbank):
    if genbank is False:
        num_fastas = len([1 for line in open(filepath_in) if line.startswith(">")])
        if meta == True:
            if num_fastas < 2:
                logger.error(
                    "ERROR: -m meta mode specified when the input file only contains 1 contig. Please re-run without specifying -m."
                )
            else:
                message = f"{num_fastas} input contigs detected."
                logger.info(message)
                if split == True:
                    message = "Split mode activtated. Separate output FASTA, gff and genbank files will be output for each contig."
                    logger.info(message)
        else:
            if num_fastas > 1:
                message = "More than one contig detected in the input file. Re-running pharokka with -m meta mode is recommended unless this is a fragmented isolate genome. Continuing."
                logger.info(message)
            if split == True:
                message = "-s or --split was specified without -m or --meta and will be ignored. Please specify -s with -m if you want to run split mode. Continuing."
                logger.info(message)


def validate_strand(strand):
    if strand != "pos" and strand != "neg":
        logger.error(
            "Error: terminase strand was incorrectly specified. It should be either 'pos' or 'neg'. Please check your input and try again. \n"
        )


def validate_terminase_start(terminase_start):
    try:
        int(terminase_start)
    except:
        logger.error(
            "Error: terminase start coordinate specified is not an integer. Please check your input and try again. \n"
        )


def validate_terminase(filepath_in, terminase_strand, terminase_start):
    if terminase_strand == "nothing":
        logger.error(
            "Error: you specified --terminase to reorient your phage to begin with the terminase large subunit, but didn't specify its strand with --terminase_strand. Please check your input and try again. \n"
        )
    if terminase_start == "nothing":
        logger.error(
            "Error: you specified --terminase to reorient your phage to begin with the terminase large subunit, but didn't specify its start coordinate with --terminase_start. Please check your input and try again. \n"
        )
    validate_strand(terminase_strand)
    validate_terminase_start(terminase_start)
    num_fastas = len([1 for line in open(filepath_in) if line.startswith(">")])
    if num_fastas > 1:
        logger.error(
            "Error: To reorient your phage genome to begin with the terminase large subunit, you can only input 1 phage genome contig. Multiple contigs were detected. Please try again. \n"
        )


def validate_threads(threads):
    try:
        x = int(threads)
        x += 1
    except:
        message = (
            "Error: you specified a non-integer value for threads of "
            + threads
            + ". Please check your input and try Pharokka again."
        )
        logger.error(message)


#######
# dependencies
#######


def check_dependencies(skip_mash):
    """Checks the dependencies and versions
    skip_mash flag from args, won't check mash is skip mash specified
    :return:
    """
    #############
    # phanotate
    #############
    try:
        process = sp.Popen(
            ["phanotate.py", "--version"], stdout=sp.PIPE, stderr=sp.DEVNULL
        )
    except:
        logger.error("Phanotate not found. Please reinstall pharokka.")
    phan_out, _ = process.communicate()
    phanotate_version = phan_out.decode().strip()

    # this handles the warning for setuptools https://github.com/gbouras13/pharokka/issues/395
    if "\n" in phanotate_version.strip():
        logger.warning("pkg_resources in phanotate is throwing a warning")
        logger.warning(
            "this should not affect pharokka but please make an issue at https://github.com/gbouras13/pharokka/issues if pharokka crashes"
        )
        logger.warning("printing the full warning for debugging if required")
        logger.warning(phanotate_version)
        phanotate_version = phanotate_version.strip().split("\n")[-1]

    phanotate_major_version = int(phanotate_version.split(".")[0])
    phanotate_minor_version = int(phanotate_version.split(".")[1])
    phanotate_minorest_version = phanotate_version.split(".")[2]

    logger.info(
        f"Phanotate version found is v{phanotate_major_version}.{phanotate_minor_version}.{phanotate_minorest_version}"
    )

    if phanotate_major_version < 1:
        logger.error("Phanotate is too old - please reinstall pharokka.")
    if phanotate_minor_version < 5:
        logger.error("Phanotate is too old - please reinstall pharokka.")

    logger.info("Phanotate version is ok.")

    #############
    # mmseqs
    #############

    try:
        process = sp.Popen(["mmseqs", "version"], stdout=sp.PIPE, stderr=sp.STDOUT)
    except:
        logger.error("MMseqs2 not found. Please reinstall pharokka.")

    mmseqs_out, _ = process.communicate()
    mmseqs_out = mmseqs_out.decode()

    mmseqs_version = mmseqs_out.strip()

    if "8cc5c" in mmseqs_version:
        mmseqs_major_version = "18"
        mmseqs_minor_version = "8cc5c"
    elif "b804f" in mmseqs_version:
        mmseqs_major_version = "17"
        mmseqs_minor_version = "b804f"
    elif "747c6" in mmseqs_version:
        mmseqs_major_version = "16"
        mmseqs_minor_version = "747c6"
    elif "6f452" in mmseqs_version:
        mmseqs_major_version = "15"
        mmseqs_minor_version = "6f452"
    elif "7e284" in mmseqs_version:
        mmseqs_major_version = "14"
        mmseqs_minor_version = "7e284"
    elif "45111" in mmseqs_version:
        mmseqs_major_version = "13"
        mmseqs_minor_version = "45111"
        logger.warning(
            f"MMseqs2 version found is v{mmseqs_major_version}.{mmseqs_minor_version}"
        )
        logger.warning(
            f"Pharokka from v1.8.0 onwards will not work with MMseqs2 versions older than v14-7e284 (October 13 2022)"
        )
        logger.error(
            f"Please install a newer version of MMseqs2, ideally v18.8cc5c or newer"
        )
    else:
        logger.warning(f"MMseqs2 version found is v{mmseqs_version}")
        logger.warning(
            f"Pharokka is recommended to be run with MMseqs2 v18.8cc5c or newer"
        )
        logger.warning(
            f"Using a different MMseqs2 version after v14-7e284 (October 13 2022) is likely to work with Pharokka without issue, but this cannot be guaranteed."
        )
        logger.warning(
            f"Using an MMseqs2 version before v14-7e284 (October 13 2022) will not work with Pharokka from v1.8.0 onwards."
        )
        logger.warning(
            f"If that is the case, please install a newer version of MMseqs2s."
        )

    logger.info(
        f"MMseqs2 version found is v{mmseqs_major_version}.{mmseqs_minor_version}"
    )
    logger.info("MMseqs2 version is ok.")

    #############
    # trnascan
    #############
    try:
        process = sp.Popen(["tRNAscan-SE", "-h"], stdout=sp.PIPE, stderr=sp.STDOUT)
    except:
        logger.error("tRNAscan-SE not found. Please reinstall pharokka.")

    trna_out, _ = process.communicate()
    trna_out = trna_out.decode()

    version_line = []

    for line in trna_out.split("\n"):
        if "tRNAscan-SE" in line:
            version_line.append(line)

    trna_version = version_line[0].split(" ")[1]
    trna_major_version = int(trna_version.split(".")[0])
    trna_minor_version = int(trna_version.split(".")[1])
    trna_minorest_version = int(trna_version.split(".")[2])

    logger.info(
        f"tRNAscan-SE version found is v{trna_major_version}.{trna_minor_version}.{trna_minorest_version}"
    )

    if trna_major_version != 2:
        logger.error("tRNAscan-SE is the wrong version. Please re-install pharokka.")
    if trna_minor_version != 0:
        logger.error("tRNAscan-SE is the wrong version. Please re-install pharokka.")
    if trna_minorest_version < 9:
        logger.error("tRNAscan-SE is the wrong version. Please re-install pharokka.")

    logger.info("tRNAscan-SE version is ok.")

    #############
    # minced
    #############
    try:
        process = sp.Popen(["minced", "--version"], stdout=sp.PIPE, stderr=sp.STDOUT)
    except:
        logger.error("MinCED not found. Please reinstall pharokka.")

    minced_out, _ = process.communicate()
    minced_out = minced_out.decode()

    version_line = []

    for line in minced_out.split("\n"):
        if "minced" in line:
            version_line.append(line)

    minced_version = version_line[0].split(" ")[1]
    minced_major_version = int(minced_version.split(".")[0])
    minced_minor_version = int(minced_version.split(".")[1])
    minced_minorest_version = int(minced_version.split(".")[2])

    logger.info(
        f"MinCED version found is v{minced_major_version}.{minced_minor_version}.{minced_minorest_version}"
    )

    if minced_major_version != 0:
        logger.error("MinCED is the wrong version. Please re-install pharokka.")
    if minced_minor_version != 4:
        logger.error("MinCED is the wrong version. Please re-install pharokka.")
    if minced_minorest_version < 2:
        logger.error("MinCED is the wrong version. Please re-install pharokka.")

    logger.info("MinCED version is ok.")

    #############
    # aragorn
    #############
    try:
        process = sp.Popen(["aragorn", "-h"], stdout=sp.PIPE, stderr=sp.STDOUT)
    except:
        logger.error("ARAGORN not found. Please reinstall pharokka.")

    aragorn_out, _ = process.communicate()
    aragorn_out = aragorn_out.decode()

    version_line = []

    for line in aragorn_out.split("\n"):
        if "Dean Laslett" in line:
            version_line.append(line)

    aragorn_version = version_line[0].split(" ")[1]
    # strip off v
    aragorn_version = aragorn_version[1:]
    aragorn_major_version = int(aragorn_version.split(".")[0])
    aragorn_minor_version = int(aragorn_version.split(".")[1])
    aragorn_minorest_version = int(aragorn_version.split(".")[2])

    logger.info(
        f"ARAGORN version found is v{aragorn_major_version}.{aragorn_minor_version}.{aragorn_minorest_version}"
    )

    if aragorn_major_version != 1:
        logger.error("ARAGORN is the wrong version. Please re-install pharokka.")
    if aragorn_minor_version != 2:
        logger.error("ARAGORN is the wrong version. Please re-install pharokka.")
    if aragorn_minorest_version < 41:
        logger.error("ARAGORN is the wrong version. Please re-install pharokka.")

    logger.info("ARAGORN version is ok.")

    #############
    # mash
    #############
    if skip_mash is False:
        try:
            process = sp.Popen(["mash", "--version"], stdout=sp.PIPE, stderr=sp.STDOUT)
        except:
            logger.error("mash not found. Please reinstall pharokka.")

        mash_out, _ = process.communicate()
        mash_out = mash_out.decode().strip()

        mash_major_version = int(mash_out.split(".")[0])
        mash_minor_version = int(mash_out.split(".")[1])

        logger.info(f"mash version found is v{mash_major_version}.{mash_minor_version}")

        if mash_major_version != 2:
            logger.error("mash is the wrong version. Please re-install pharokka.")
        if mash_minor_version < 2:
            logger.error("mash is the wrong version. Please re-install pharokka.")

        logger.info("mash version is ok.")

    #############
    # dnaapler
    #############
    try:
        process = sp.Popen(["dnaapler", "--version"], stdout=sp.PIPE, stderr=sp.STDOUT)
    except:
        logger.error("Dnaapler not found. Please reinstall pharokka.")

    dnaapler_out, _ = process.communicate()
    dnaapler_out = dnaapler_out.decode()

    dnaapler_out = dnaapler_out.split("n ")[1]

    dnaapler_major_version = int(dnaapler_out.split(".")[0])
    dnaapler_minor_version = int(dnaapler_out.split(".")[1])
    dnaapler_minorest_version = int(dnaapler_out.split(".")[2])

    logger.info(
        f"Dnaapler version found is v{dnaapler_major_version}.{dnaapler_minor_version}.{dnaapler_minorest_version}"
    )

    if dnaapler_major_version < 1:
        logger.error(
            "Dnaapler is the wrong version. Please install Dnaapler v1.0.1 or higher."
        )

    logger.info("Dnaapler version is ok.")

    #######
    # pyrodigal
    #######

    pyrodigal_version = pyrodigal.__version__
    pyrodigal_major_version = int(pyrodigal_version.split(".")[0])

    if pyrodigal_major_version < 3:
        logger.error("Pyrodigal is the wrong version. Please re-install pharokka.")

    logger.info(f"Pyrodigal version is v{pyrodigal_version}")
    logger.info(f"Pyrodigal version is ok.")

    #######
    # pyrodigal gv
    #######

    pyrodigal_gv_version = pyrodigal_gv.__version__
    pyrodigal_major_version = int(pyrodigal_gv_version.split(".")[0])

    if pyrodigal_major_version < 0:
        logger.error("Pyrodigal_gv is the wrong version. Please re-install pharokka.")

    logger.info(f"Pyrodigal_gv version is v{pyrodigal_gv_version}")
    logger.info(f"Pyrodigal_gv version is ok.")

    #######
    # pyrodigal rv
    #######

    pyrodigal_rv_version = pyrodigal_rv.__version__
    pyrodigal_major_version = int(pyrodigal_rv_version.split(".")[0])

    if pyrodigal_major_version < 0:
        logger.error("Pyrodigal_rv is the wrong version. Please re-install pharokka.")

    logger.info(f"Pyrodigal_rv version is v{pyrodigal_rv_version}")
    logger.info(f"Pyrodigal_rv version is ok.")


    return (
        phanotate_version,
        pyrodigal_version,
        pyrodigal_gv_version,
        pyrodigal_rv_version,
        trna_version,
        aragorn_version,
        minced_version,
    )


def instantiate_split_output(out_dir, split):
    """
    if the split flag is true, will create these output directories
    """
    if split == True:
        single_gff_dir = os.path.join(out_dir, "single_gffs")
        single_gbk_dir = os.path.join(out_dir, "single_gbks")

        if os.path.isdir(single_gff_dir) == False:
            os.mkdir(single_gff_dir)
        if os.path.isdir(single_gbk_dir) == False:
            os.mkdir(single_gbk_dir)


def validate_custom_hmm(filename):
    suffix = ".h3m"

    logger.info(f"Checking custom hmm profile {filename}")

    if filename.endswith(suffix) is True:
        logger.info(f"{filename} checked.")
    else:
        logger.error(
            f"{filename} does not end with .h3m . Please check your --custom_hmm parameter or use create_custom_hmm.py to create a custom HMM profile."
        )


def validate_and_extract_genbank(filename, out_dir):
    # Open the GenBank file
    try:
        with open(filename, "r") as genbank_file:
            for record in SeqIO.parse(genbank_file, "genbank"):
                output_fasta_path = f"{out_dir}/genbank.fasta"

                # Write the sequence in FASTA format to the output file
                with open(output_fasta_path, "a") as output_fasta:
                    SeqIO.write(record, output_fasta, "fasta")
    except:
        logger.error(f"{filename} is not a genbank format file.")



#!/usr/bin/env python3
import argparse
import os
import shutil
import sys
from argparse import RawTextHelpFormatter
from pathlib import Path

from loguru import logger
from plot import create_single_plot
from pycirclize.parser import Genbank
from util import get_version


def get_input():
    """gets input for pharokka_plotter.py
    :return: args
    """
    parser = argparse.ArgumentParser(
        description="pharokka_multiplotter.py: pharokka plotting function for muliple phages",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "-g",
        "--genbank",
        action="store",
        required=True,
        help="Input genbank file from Pharokka.",
    )
    parser.add_argument(
        "-o",
        "--outdir",
        action="store",
        help="Pharokka output directory.",
        required=True,
    )
    parser.add_argument(
        "-f", "--force", help="Overwrites the output plot file.", action="store_true"
    )
    parser.add_argument(
        "--label_hypotheticals",
        help="Flag to label  hypothetical or unknown proteins. By default these are not labelled.",
        action="store_true",
    )
    parser.add_argument(
        "--remove_other_features_labels",
        help="Flag to remove labels for tRNA/tmRNA/CRISPRs. By default these are labelled. \nThey will still be plotted in black.",
        action="store_true",
    )
    parser.add_argument(
        "--title_size",
        action="store",
        default="20",
        help="Controls title size. Must be an integer. Defaults to 20.",
    )
    parser.add_argument(
        "--label_size",
        action="store",
        default="8",
        help="Controls annotation label size. Must be an integer. Defaults to 8.",
    )
    parser.add_argument(
        "--interval",
        action="store",
        default="5000",
        help="Axis tick interval. Must be an integer. Must be an integer. Defaults to 5000.",
    )
    parser.add_argument(
        "--truncate",
        action="store",
        default="20",
        help="Number of characters to include in annoation labels before truncation with ellipsis. \nMust be an integer. Defaults to 20.",
    )
    parser.add_argument(
        "--dpi",
        action="store",
        default="600",
        help="Resultion (dots per inch). Must be an integer. Defaults to 600.",
    )
    parser.add_argument(
        "--annotations",
        action="store",
        default="1",
        help="Controls the proporition of annotations labelled. Must be a number between 0 and 1 inclusive. \n0 = no annotations, 0.5 = half of the annotations, 1 = all annotations. \nDefaults to 1. Chosen in order of CDS size.",
    )
    parser.add_argument(
        "-t", "--plot_title", action="store", default="Phage", help="Plot name."
    )
    parser.add_argument(
        "--label_ids",
        action="store",
        default="",
        help="Text file with list of CDS IDs (from gff file) that are guaranteed to be labelled.",
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = get_input()
    # preamble
    logger.add(lambda _: sys.exit(1), level="ERROR")
    logger.info(f"Starting Pharokka v{get_version()}")
    logger.info("Running pharokka_multiplotter.py to plot your phages.")
    logger.info("Command executed: {}", args)
    logger.info("Repository homepage is https://github.com/gbouras13/pharokka")
    logger.info("Written by George Bouras: george.bouras@adelaide.edu.au")
    logger.info("Checking your inputs.")

    try:
        int(args.interval)
    except:
        logger.error(
            f"--interval {args.interval} specified is not an integer. Please check your input and try again."
        )

    try:
        int(args.label_size)
    except:
        logger.error(
            f"--label_size {args.label_size} specified is not an integer. Please check your input and try again."
        )

    try:
        int(args.title_size)
    except:
        logger.error(
            f"--title_size {args.title_size} specified is not an integer. Please check your input and try again."
        )

    try:
        int(args.dpi)
    except:
        logger.error(
            f"--dpi {args.dpi} specified is not an integer. Please check your input and try again."
        )

    try:
        float(args.annotations)
    except:
        logger.error(
            f"--annotations {args.annotations} specified is not a float. Please check your input and try again."
        )

    if args.force == True:
        if os.path.exists(args.outdir) == True:
            logger.info(f"Removing {args.outdir} as --force was specified.")
            shutil.rmtree(args.outdir)
        else:
            logger.warning(
                f"--force was specified even though the output plot directory {args.outdir} does not already exist."
            )
            logger.warning("Continuing")
    else:
        if os.path.exists(args.outdir) == True:
            logger.error(
                f"Output directoey {args.outdir} already exists and force was not specified. Please specify -f or --force to overwrite the output directory."
            )

    # instantiate outdir
    if Path(args.outdir).exists() is False:
        Path(args.outdir).mkdir(parents=True, exist_ok=True)

    # check label_ids

    # list of all IDs that need to be labelled from file
    label_force_list = []

    if args.label_ids != "":
        logger.info(
            f"You have specified a file {args.label_ids} containing a list of CDS IDs to force label."
        )
        # check if it is a file
        if os.path.isfile(args.label_ids) == False:
            logger.error(f"{args.label_ids} was not found.")
        # check if it contains text
        try:
            # Open the file in read mode
            with open(Path(args.label_ids), "r") as file:
                # Read the first character
                first_char = file.read(1)

                # read all the labels
                with open(Path(args.label_ids)) as f:
                    ignore_dict = {x.rstrip().split()[0] for x in f}
                # label force list
                label_force_list = list(ignore_dict)

        except FileNotFoundError:
            logger.warning(
                f"{args.label_id} contains no text. No contigs will be ignored"
            )

    logger.info("All files checked.")

    # single threaded plots
    threads = 1

    gbk = Genbank(args.genbank)

    # gets all contigs and seqs
    gb_seq_dict = gbk.get_seqid2seq()

    gb_size_dict = gbk.get_seqid2size()

    contig_count = len(gb_seq_dict)

    # gets all features - will get all regardless of type (tRNA etc from pharokka)
    gb_feature_dict = gbk.get_seqid2features()

    # check label_ids
    # list of all IDs that need to be labelled from file
    label_force_list = []

    label_ids = args.label_ids

    if label_ids != "":
        logger.info(
            f"You have specified a file {label_ids} containing a list of CDS IDs to force label."
        )
        # check if it is a file
        if Path(label_ids).exists() is False:
            logger.error(f"{label_ids} was not found.")
        # check if it contains text
        try:
            # Open the file in read mode
            with open(Path(label_ids), "r") as file:
                # Read the first character
                # will error if file is empty
                first_char = file.read(1)

                # read all the labels
                with open(Path(label_ids)) as f:
                    ignore_dict = {x.rstrip().split()[0] for x in f}
                # label force list
                label_force_list = list(ignore_dict)

        except FileNotFoundError:
            logger.warning(f"{label_ids} contains no text. No contigs will be ignored")

    # if there is 1 contig, then all the parameters will apply

    for contig_id, contig_sequence in gb_seq_dict.items():
        logger.info(f"Plotting {contig_id}")

        create_single_plot(
            contig_id,
            contig_sequence,
            contig_count,
            gb_size_dict,
            gb_feature_dict,
            gbk,
            int(args.interval),
            float(args.annotations),
            int(args.title_size),
            args.plot_title,
            int(args.truncate),
            args.outdir,
            int(args.dpi),
            float(args.label_size),
            args.label_hypotheticals,
            args.remove_other_features_labels,
            label_force_list,
        )



"""
Holds functions for pharokka_proteins.py
"""

import argparse
import collections
import os
from argparse import RawTextHelpFormatter
from cmath import nan
from pathlib import Path


import numpy as np
import pandas as pd
import pyhmmer
from Bio import SeqIO
from external_tools import ExternalTool
from loguru import logger
from post_processing import (process_card_results, process_pyhmmer_results,
                             process_vfdb_results)
from pyhmmer.easel import SequenceFile
from pyhmmer.plan7 import HMM, HMMFile
from util import ( get_contig_headers, get_version,
                  remove_directory)

Result = collections.namedtuple("Result", ["protein", "phrog", "bitscore", "evalue"])


pd.options.mode.chained_assignment = None


def get_input_proteins():
    parser = argparse.ArgumentParser(
        description="pharokka_proteins.py: fast phage protein annotation script",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--infile",
        action="store",
        help="Input proteins file in amino acid fasta format.",
    )
    parser.add_argument(
        "-o",
        "--outdir",
        action="store",
        help="Directory to write the output to.",
        default=os.path.join(os.getcwd(), "output/"),
    )
    parser.add_argument(
        "-d",
        "--database",
        action="store",
        help="Database directory. If the databases have been installed in the default directory, this is not required. Otherwise specify the path.",
        default="Default",
    )
    parser.add_argument(
        "-t",
        "--threads",
        help="Number of threads. Defaults to 1.",
        action="store",
        default=str(1),
    )
    parser.add_argument(
        "-f", "--force", help="Overwrites the output directory.", action="store_true"
    )
    parser.add_argument(
        "-p",
        "--prefix",
        action="store",
        help="Prefix for output files. This is not required.",
        default="Default",
    )
    parser.add_argument(
        "-e",
        "--evalue",
        help="E-value threshold for MMseqs2 database PHROGs, VFDB and CARD and pyhmmer PHROGs database search. Defaults to 1E-05.",
        action="store",
        default="1E-05",
    )
    parser.add_argument(
        "--hmm_only",
        help="Runs pyhmmer (HMMs) with PHROGs only, not MMseqs2 with PHROGs, CARD or VFDB.",
        action="store_true",
    )
    parser.add_argument(
        "--mmseqs2_only",
        help="Runs MMseqs2 with PHROGs, CARD and VFDB only (same as Pharokka v1.3.2 and prior).",
        action="store_true",
    )
    parser.add_argument(
        "--reverse_mmseqs2",
        help="MMseqs2 database as target not query.",
        action="store_true",
    )
    parser.add_argument(
        "--sensitivity",
        help="MMseqs2 sensitivity.",
        default=8.5,
        type=float,
    )
    parser.add_argument(
        "-V",
        "--version",
        help="Print pharokka Version",
        action="version",
        version=get_version(),
    )
    parser.add_argument(
        "--citation", help="Print pharokka Citation", action="store_true"
    )
    args = parser.parse_args()

    return args


def run_mmseqs_proteins(input_fasta, db_dir, out_dir, threads, logdir, evalue, reverse_mmseqs2, sensitivity, db_name):
    """
    Runs mmseqs2 for pharokka_proteins.py
    :param db_dir: database path
    :param out_dir: output directory
    :param logger: logger
    :params threads: threads
    :param gene_predictor: phanotate or prodigal
    :param evalue: evalue for mmseqs2
    :param reverse_mmseqs2: reverse mmseqs2 database to be target
    :param sensitivity: MMseqs2 sensitivity
    :param db_name: str one of 'PHROG', 'VFDB' or 'CARD'
    :return:
    """

    logger.info(f"Running MMseqs2 on {db_name} Database.")

    # define the outputs
    if db_name == "PHROG":
        mmseqs_dir = os.path.join(out_dir, "mmseqs/")
        target_db_dir = os.path.join(out_dir, "target_dir/")
        tmp_dir = os.path.join(out_dir, "tmp_dir/")
        profile_db = os.path.join(db_dir, "phrogs_profile_db")
        mmseqs_result_tsv = os.path.join(out_dir, "mmseqs_results.tsv")
    elif db_name == "VFDB":
        mmseqs_dir = os.path.join(out_dir, "VFDB/")
        target_db_dir = os.path.join(out_dir, "VFDB_target_dir/")
        tmp_dir = os.path.join(out_dir, "VFDB_dir/")
        profile_db = os.path.join(db_dir, "vfdb")
        mmseqs_result_tsv = os.path.join(out_dir, "vfdb_results.tsv")
    elif db_name == "CARD":
        mmseqs_dir = os.path.join(out_dir, "CARD/")
        target_db_dir = os.path.join(out_dir, "CARD_target_dir/")
        tmp_dir = os.path.join(out_dir, "VFDB_dir/")
        profile_db = os.path.join(db_dir, "CARD")
        mmseqs_result_tsv = os.path.join(out_dir, "CARD_results.tsv")

    target_seqs = os.path.join(target_db_dir, "target_seqs")

    # make dir for target db
    if os.path.isdir(target_db_dir) == False:
        os.mkdir(target_db_dir)

    # creates db for input
    mmseqs_createdb = ExternalTool(
        tool="mmseqs createdb",
        input=f"",
        output=f"{target_seqs}",
        params=f"{input_fasta}",  # param goes before output and mmseqs2 required order
        logdir=logdir,
        outfile="",
    )

    ExternalTool.run_tool(mmseqs_createdb)

    # runs the mmseqs seacrh

    result_mmseqs = os.path.join(mmseqs_dir, "results_mmseqs")

    if db_name == "PHROG":

        if reverse_mmseqs2:
            params_list = f"-e {evalue} {target_seqs} {profile_db} {result_mmseqs}"
        else:
            params_list = f"-e {evalue} {profile_db} {target_seqs} {result_mmseqs}"

        mmseqs_search = ExternalTool(
            tool="mmseqs search",
            input=f"",
            output=f"{tmp_dir} -s {sensitivity} --threads {threads}",
            params=params_list,  # param goes before output and mmseqs2 required order
            logdir=logdir,
            outfile="",
        )
    else:  # if it is vfdb or card search with cutoffs instead of evalue

        if reverse_mmseqs2:
            params_list = f"--min-seq-id 0.8 -c 0.4 {target_seqs} {profile_db} {result_mmseqs}"
        else:
            params_list = f"--min-seq-id 0.8 -c 0.4 {profile_db} {target_seqs} {result_mmseqs}"

        mmseqs_search = ExternalTool(
            tool="mmseqs search",
            input=f"",
            output=f"{tmp_dir} -s {sensitivity} --threads {threads}",
            params=params_list,  # param goes before output and mmseqs2 required order
            logdir=logdir,
            outfile="",
        )

    ExternalTool.run_tool(mmseqs_search)


    if reverse_mmseqs2:
        params_list = f"{target_seqs} {profile_db} {result_mmseqs}"
    else:
        params_list = f"{profile_db} {target_seqs} {result_mmseqs}" # param goes before output and mmseqs2 required order


    # creates the output tsv
    mmseqs_createtsv = ExternalTool(
        tool="mmseqs createtsv",
        input=f"",
        output=f"{mmseqs_result_tsv} --full-header --threads {threads}",
        params=params_list,  # param goes before output and mmseqs2 required order
        logdir=logdir,
        outfile="",
    )

    ExternalTool.run_tool(mmseqs_createtsv)

    # remove the target dir when finished
    remove_directory(target_db_dir)


# from itertools import chain


def run_pyhmmer_proteins(input_fasta, db_dir, threads, evalue):
    """
    Runs phymmer on phrogs
    :param input_fasta: input FASTA
    :param db_dir: database path
    :params threads: threads
    :param evalue: evalue threshold for pyhmmer
    :return: best_results - dictionary of top HMM hits for each protein
    """

    # define result
    Result = collections.namedtuple(
        "Result", ["protein", "phrog", "bitscore", "evalue"]
    )

    # run hmmscan and get all results
    results = []
    alphabet = (
        pyhmmer.easel.Alphabet.amino()
    )  # https://github.com/althonos/pyhmmer/issues/80 to solve #357 #331 need to specify the alphabet explicitly
    with pyhmmer.plan7.HMMFile(os.path.join(db_dir, "all_phrogs.h3m")) as hmms:  # hmms
        with pyhmmer.easel.SequenceFile(
            input_fasta, digital=True, alphabet=alphabet
        ) as seqs:  # amino acid sequences
            for hits in pyhmmer.hmmer.hmmscan(
                seqs, hmms, cpus=int(threads), E=float(evalue)
            ):  # run hmmscan
                protein = (
                    hits.query.name.decode()
                )  # get protein from the hit query.name - this changed in pyhmmer v 0.11.0 from hits.query_name.decode() which was removed. So need pyhmmer >=0.11.0
                for hit in hits:
                    if hit.included:
                        # include the hit to the result collection
                        results.append(
                            Result(protein, hit.name.decode(), hit.score, hit.evalue)
                        )

    # get  best results for each protein
    best_results = {}
    keep_protein = set()
    for result in results:
        if result.protein in best_results:
            previous_bitscore = best_results[result.protein].bitscore
            if result.bitscore > previous_bitscore:
                best_results[result.protein] = result
                keep_protein.add(result.protein)
            elif result.bitscore == previous_bitscore:
                if best_results[result.protein].phrog != hit.name.decode():
                    keep_protein.remove(result.protein)
        else:
            best_results[result.protein] = result
            keep_protein.add(result.protein)

    return best_results


class Pharok_Prot:
    """pharokka_proteins.py Output Class"""

    def __init__(
        self,
        out_dir: Path = "output/",
        db_dir: Path = "db_dir/",
        prefix: str = "pharokka",
        input_fasta: Path = "input.fasta",
        pyhmmer_results_dict: dict = {
            "p1": Result(protein="p1", phrog="phrog_1", bitscore=1, evalue=2.01e-01)
        },
        tophits_df: pd.DataFrame() = pd.DataFrame(
            {"col1": [1, 2, 3], "col2": [4, 5, 6]}
        ),
        vfdb_results: pd.DataFrame() = pd.DataFrame(
            {"col1": [1, 2, 3], "col2": [4, 5, 6]}
        ),
        card_results: pd.DataFrame() = pd.DataFrame(
            {"col1": [1, 2, 3], "col2": [4, 5, 6]}
        ),
        length_df: pd.DataFrame() = pd.DataFrame(
            {"col1": [1, 2, 3], "col2": [4, 5, 6]}
        ),
        mmseqs_flag: bool = True,
        hmm_flag: bool = True,
        reverse_mmseqs2: bool = False
    ) -> None: 
        """
        Parameters
        --------
        out_dir: str,
            output directory
        db_dir: str,
            database directory
        prefix: str, prefix for output
            prefix
        input_fasta: Path
            input FASTA file with phage contigs
        pyhmmer_results_dict: dict
            dictionary of Result tuples from pyhmmer.
        tophits_df: pd.DataFrame, required
            merged dataframe output
        vfdb_results: pd.DataFrame,
            vfdb dataframe output
        card_results: pd.DataFrame,
            CARD dataframe output
        length_df: pd.DataFrame,
            dataframe with lengths for each input contig
        run_mmseqs: bool
            whether MMseqs2 was run
        run_hmm: bool
            whether HMM was run
        reverse_mmseqs2: bool
            whether to run MMseqs using database as target
        """
        self.out_dir = out_dir
        self.db_dir = db_dir
        self.prefix = prefix
        self.input_fasta = input_fasta
        self.pyhmmer_results_dict = pyhmmer_results_dict
        self.tophits_df = tophits_df
        self.vfdb_results = vfdb_results
        self.card_results = card_results
        self.length_df = length_df
        self.mmseqs_flag = mmseqs_flag
        self.hmm_flag = hmm_flag
        self.reverse_mmseqs2 = reverse_mmseqs2

    def process_dataframes(self):
        """
        Processes and combines PHROGS, CARD and VFDB mmseqs output
        :param self: Pharok class object
        :param db_dir: database directory path
        :param out_dir: output directory path
        :return:
        """

        ####### ####### ####### ####### #######
        ####### ####### ####### ####### #######
        # get tophits 
        ####### ####### ####### ####### #######
        ####### ####### ####### ####### #######

        ids = get_contig_headers(self.input_fasta) # returns the seq.description
        all_genes_df = pd.DataFrame({"gene": ids})
        
        if self.mmseqs_flag is True: 
            # dummy df of all genes

            # MMseqs PHROGs file
            mmseqs_file = os.path.join(self.out_dir, "mmseqs_results.tsv")
            logger.info("Processing mmseqs2 output.")

            if self.reverse_mmseqs2:
                col_list = [
                "gene",
                "mmseqs_phrog",
                "mmseqs_alnScore",
                "mmseqs_seqIdentity",
                "mmseqs_eVal",
                "qStart",
                "qEnd",
                "qLen",
                "tStart",
                "tEnd",
                "tLen"]

            else:
                col_list = [
                "mmseqs_phrog",
                "gene",
                "mmseqs_alnScore",
                "mmseqs_seqIdentity",
                "mmseqs_eVal",
                "qStart",
                "qEnd",
                "qLen",
                "tStart",
                "tEnd",
                "tLen"]


            mmseqs_df = pd.read_csv(
                mmseqs_file, delimiter="\t", index_col=False, names=col_list
            )

             # get best hit per gene (lowest e-value)
            tophits_df = (
                mmseqs_df
                .sort_values("mmseqs_eVal")
                .groupby("gene", as_index=False)
                .first()
                [["gene", "mmseqs_phrog", "mmseqs_alnScore",
                "mmseqs_seqIdentity", "mmseqs_eVal"]]
            )

            # get all genes by leftjoining with the all gene df

            tophits_df = (
                all_genes_df
                .merge(tophits_df, on="gene", how="left")
                .fillna({
                    "mmseqs_phrog": "No_MMseqs",
                    "mmseqs_alnScore": "No_MMseqs",
                    "mmseqs_seqIdentity": "No_MMseqs",
                    "mmseqs_eVal": "No_MMseqs",
                })
            )

        # tophits for pyhmmer only
        else:
            # create tophits df
            tophits_df = all_genes_df.assign(
                mmseqs_phrog="No_MMseqs",
                mmseqs_alnScore="No_MMseqs",
                mmseqs_seqIdentity="No_MMseqs",
                mmseqs_eVal="No_MMseqs",
            )



        ####################
        # combine phrogs
        ####################

        # Adds pyhmmer results if tru
        if self.hmm_flag is True:
            tophits_df = process_pyhmmer_results(tophits_df, self.pyhmmer_results_dict)
        else:
            tophits_df["pyhmmer_phrog"] = "No_HMM"
            tophits_df["pyhmmer_bitscore"] = "No_HMM"
            tophits_df["pyhmmer_evalue"] = "No_HMM"

        # strip off phrog_ for both
        tophits_df["mmseqs_phrog"] = tophits_df["mmseqs_phrog"].str.replace(
            "phrog_", ""
        )
        tophits_df["pyhmmer_phrog"] = tophits_df["pyhmmer_phrog"].str.replace(
            "phrog_", ""
        )

        ############
        # code to create 1 overall phrog column
        # pick the mmseqs PHROG column first if it was run

        if self.mmseqs_flag is True:
            tophits_df["phrog"] = tophits_df["mmseqs_phrog"]
            # add pyhmmer phrog for any entry without mmseqs
            for index, row in tophits_df.iterrows():
                if isinstance(
                    row["phrog"], float
                ):  # for all the rows without an mmseqs2 PHROG will be nan - floats
                    # to write all hits where there was no mmseqs but there was a hmm
                    if row["pyhmmer_phrog"] != "No_PHROG":
                        tophits_df.at[index, "phrog"] = row["pyhmmer_phrog"]

        else:  # only need to worry about pyhmmer
            tophits_df["phrog"] = tophits_df["pyhmmer_phrog"]

        # read in phrog annotaion file
        phrog_annot_df = pd.read_csv(
            os.path.join(self.db_dir, "phrog_annot_v4.tsv"), sep="\t", index_col=False
        )
        phrog_annot_df["phrog"] = phrog_annot_df["phrog"].astype(str)

        # merge phrog
        # get only the contig id not the full description from mmseqs2 output
        tophits_df["gene"] = tophits_df["gene"].str.split(" ").str[0]

        tophits_df = tophits_df.merge(phrog_annot_df, on="phrog", how="left")
        # convert no phrog to hyp protein
        tophits_df["annot"] = tophits_df["annot"].str.replace(
            "No_PHROG", "hypothetical protein"
        )
        tophits_df["category"] = tophits_df["category"].str.replace(
            "No_PHROG", "unknown function"
        )

        # # replace with No_PHROG if nothing found
        # cast columns to str
        tophits_df["mmseqs_phrog"] = tophits_df["mmseqs_phrog"].astype(str)
        tophits_df["mmseqs_alnScore"] = tophits_df["mmseqs_alnScore"].astype(str)
        tophits_df["mmseqs_seqIdentity"] = tophits_df["mmseqs_seqIdentity"].astype(str)
        tophits_df["mmseqs_eVal"] = tophits_df["mmseqs_eVal"].astype(str)
        tophits_df["color"] = tophits_df["color"].astype(str)
        tophits_df = tophits_df.replace(np.nan, "No_PHROG", regex=True)
        tophits_df.loc[tophits_df["mmseqs_phrog"] == "No_PHROG", "mmseqs_phrog"] = (
            "No_PHROG"
        )
        tophits_df.loc[
            tophits_df["mmseqs_alnScore"] == "No_PHROG", "mmseqs_alnScore"
        ] = "No_PHROG"
        tophits_df.loc[
            tophits_df["mmseqs_seqIdentity"] == "No_PHROG", "mmseqs_seqIdentity"
        ] = "No_PHROG"

        tophits_df.loc[
            tophits_df["mmseqs_eVal"] == "No_PHROG", "mmseqs_eVal"
        ] = "No_PHROG"

        tophits_df.loc[tophits_df["color"] == "No_PHROG", "color"] = "No_PHROG"

        # get phrog
        tophits_df["phrog"] = tophits_df["phrog"].astype(str)
        phrog_annot_df["phrog"] = phrog_annot_df["phrog"].astype(str)

        # drop existing color annot category cols and
        tophits_df = tophits_df.drop(columns=["color", "annot", "category"])
        tophits_df = tophits_df.merge(phrog_annot_df, on="phrog", how="left")
        tophits_df["annot"] = tophits_df["annot"].replace(
            np.nan, "hypothetical protein", regex=True
        )
        tophits_df["category"] = tophits_df["category"].replace(
            np.nan, "unknown function", regex=True
        )
        tophits_df["color"] = tophits_df["color"].replace(nan, "None", regex=True)

        ##### length df

        fasta_sequences = SeqIO.parse(open(self.input_fasta), "fasta")
        contig_names = []
        lengths = []

        for fasta in fasta_sequences:
            contig_names.append(fasta.id)
            lengths.append(len(fasta.seq))

        length_df = pd.DataFrame(
            {
                "gene": contig_names,
                "length": lengths,
            }
        )
        self.length_df = length_df

        # convert nas to things

        tophits_df = length_df.merge(tophits_df, on="gene", how="left")

        # process vfdb results
        # handles empty files without a problem
        (tophits_df, vfdb_results) = process_vfdb_results(
            self.out_dir, tophits_df, proteins_flag=True, reverse_mmseqs2=self.reverse_mmseqs2
        )
        # process CARD results
        (tophits_df, card_results) = process_card_results(
            self.out_dir, tophits_df, self.db_dir, proteins_flag=True, reverse_mmseqs2=self.reverse_mmseqs2
        )

        # Rename the "gene" column to "id"
        tophits_df.rename(columns={"gene": "ID"}, inplace=True)

        desired_order = [
            col
            for col in tophits_df.columns
            if col not in ["phrog", "annot", "category"]
        ]
        desired_order.insert(2, "phrog")
        desired_order.insert(3, "annot")
        desired_order.insert(4, "category")

        # Creating a new DataFrame with columns rearranged
        tophits_df = tophits_df[desired_order]

        tophits_df["phrog"] = tophits_df["phrog"].fillna("No_PHROG")
        tophits_df["annot"] = tophits_df["annot"].fillna("hypothetical protein")
        tophits_df["category"] = tophits_df["category"].fillna("unknown function")
        tophits_df["mmseqs_phrog"] = tophits_df["mmseqs_phrog"].fillna("No_MMseqs")
        tophits_df["mmseqs_alnScore"] = tophits_df["mmseqs_alnScore"].fillna("No_MMseqs")
        tophits_df["mmseqs_seqIdentity"] = tophits_df["mmseqs_seqIdentity"].fillna("No_MMseqs")
        tophits_df["mmseqs_eVal"] = tophits_df["mmseqs_eVal"].fillna("No_MMseqs")
        tophits_df["pyhmmer_phrog"] = tophits_df["pyhmmer_phrog"].fillna("No_PHROGs_HMM")
        tophits_df["pyhmmer_bitscore"] = tophits_df["pyhmmer_bitscore"].fillna("No_PHROGs_HMM")
        tophits_df["pyhmmer_evalue"] = tophits_df["pyhmmer_evalue"].fillna("No_PHROGs_HMM")
        tophits_df["color"] = tophits_df["color"].fillna("None")

        tophits_df.to_csv(
            os.path.join(self.out_dir, f"{self.prefix}_full_merged_output.tsv"),
            sep="\t",
            index=False,
        )

        # save results
        self.tophits_df = tophits_df
        self.vfdb_results = vfdb_results
        self.card_results = card_results

        summary_df = self.tophits_df[["ID", "length", "phrog", "annot", "category"]]

        # .tsv file with all data full_merged_output
        summary_df.to_csv(
            os.path.join(self.out_dir, f"{self.prefix}_summary_output.tsv"),
            sep="\t",
            index=False,
        )

    def update_fasta_headers(self):
        """
        updates FASTA header with description
        """
        self.input_fasta

        # define outputs
        fasta_output_aas = os.path.join(self.out_dir, f"{self.prefix}.faa")

        # amino acids

        with open(fasta_output_aas, "w") as aa_fa:
            i = 0
            for aa_record in SeqIO.parse(self.input_fasta, "fasta"):
                aa_record.description = (
                    str(aa_record.description)
                    + " "
                    + str(self.tophits_df["annot"].iloc[i])
                )
                SeqIO.write(aa_record, aa_fa, "fasta")
                i += 1
